# Copyright (C) 2011  Chris Lalancette <clalance@redhat.com>
# Copyright (C) 2012-2017  Chris Lalancette <clalancette@gmail.com>

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

"""
Debian installation
"""

import gzip
import os
import re
import shutil
import textwrap

import oz.GuestFSManager
import oz.Linux
import oz.OzException
import oz.ozutil


class DebianConfiguration(object):
    """
    Configuration class for Debian installation.
    """
    def __init__(self, need_auto_direct, need_auto_iso, default_netdev,
                 default_diskbus):
        self._need_auto_direct = need_auto_direct
        self._need_auto_iso = need_auto_iso
        self._default_netdev = default_netdev
        self._default_diskbus = default_diskbus

    @property
    def need_auto_direct(self):
        """
        Property method for whether this version of Debian needs 'auto' on the
        command-line for direct installs.
        """
        return self._need_auto_direct

    @property
    def need_auto_iso(self):
        """
        Property method for whether this version of Debian needs 'auto' on the
        command-line for ISO installs.
        """
        return self._need_auto_iso

    @property
    def default_netdev(self):
        """
        Property method for the default netdev for this version of Debian.
        """
        return self._default_netdev

    @property
    def default_diskbus(self):
        """
        Property method for the default diskbus for this version of Debian.
        """
        return self._default_diskbus


version_to_config = {
    '9': DebianConfiguration(need_auto_direct=False, need_auto_iso=True,
                             default_netdev='virtio', default_diskbus='virtio'),
    '8': DebianConfiguration(need_auto_direct=False, need_auto_iso=True,
                             default_netdev='virtio', default_diskbus='virtio'),
    '7': DebianConfiguration(need_auto_direct=True, need_auto_iso=True,
                             default_netdev='virtio', default_diskbus='virtio'),
    '6': DebianConfiguration(need_auto_direct=False, need_auto_iso=False,
                             default_netdev='virtio', default_diskbus='virtio'),
    '5': DebianConfiguration(need_auto_direct=False, need_auto_iso=False,
                             default_netdev='virtio', default_diskbus='virtio'),
}


class DebianGuest(oz.Linux.LinuxCDGuest):
    """
    Class for Debian 5, 6, 7, 8, and 9 installation.
    """
    def __init__(self, tdl, config, auto, output_disk, netdev, diskbus,
                 macaddress):
        self.config = version_to_config[tdl.update]

        if netdev is None:
            netdev = self.config.default_netdev
        if diskbus is None:
            diskbus = self.config.default_diskbus
        oz.Linux.LinuxCDGuest.__init__(self, tdl, config, auto, output_disk,
                                       netdev, diskbus, True, True, macaddress)

        self.crond_was_active = False
        self.sshd_was_active = False
        self.sshd_config = textwrap.dedent(
            """\
            SyslogFacility AUTHPRIV
            PasswordAuthentication yes
            ChallengeResponseAuthentication no
            GSSAPIAuthentication yes
            GSSAPICleanupCredentials yes
            UsePAM yes
            AcceptEnv LANG LC_CTYPE LC_NUMERIC LC_TIME LC_COLLATE LC_MONETARY LC_MESSAGES
            AcceptEnv LC_PAPER LC_NAME LC_ADDRESS LC_TELEPHONE LC_MEASUREMENT
            AcceptEnv LC_IDENTIFICATION LC_ALL LANGUAGE
            AcceptEnv XMODIFIERS
            X11Forwarding yes
            Subsystem       sftp    /usr/libexec/openssh/sftp-server
            """
        )

        self.tunnels = {}

        self.ssh_startuplink = None
        self.cron_startuplink = None

        self.debarch = self.tdl.arch
        if self.debarch == "x86_64":
            self.debarch = "amd64"

        self.kernelfname = os.path.join(self.output_dir,
                                        self.tdl.name + "-kernel")
        self.initrdfname = os.path.join(self.output_dir,
                                        self.tdl.name + "-ramdisk")
        self.kernelcache = os.path.join(self.data_dir, "kernels",
                                        self.tdl.distro + self.tdl.update + self.tdl.arch + "-kernel")
        self.initrdcache = os.path.join(self.data_dir, "kernels",
                                        self.tdl.distro + self.tdl.update + self.tdl.arch + "-ramdisk")

    def _copy_preseed(self, outname):
        """
        Method to copy and modify an Debian style preseed file.
        """
        self.log.debug("Putting the preseed file in place")

        if self.default_auto_file():
            def _preseed_sub(line):
                """
                Method that is called back from oz.ozutil.copy_modify_file() to
                modify preseed files as appropriate for Debian.
                """
                if re.match('d-i passwd/root-password password', line):
                    return 'd-i passwd/root-password password ' + self.rootpw + '\n'
                elif re.match('d-i passwd/root-password-again password', line):
                    return 'd-i passwd/root-password-again password ' + self.rootpw + '\n'
                return line

            oz.ozutil.copy_modify_file(self.auto, outname, _preseed_sub)
        else:
            shutil.copy(self.auto, outname)

    def _modify_iso(self):
        """
        Method to make the boot ISO auto-boot with appropriate parameters.
        """
        self.log.debug("Modifying ISO")
        self.log.debug("Copying preseed file")
        outname = os.path.join(self.iso_contents, "preseed", "customiso.seed")
        outdir = os.path.dirname(outname)
        oz.ozutil.mkdir_p(outdir)
        self._copy_preseed(outname)

        # arch == i386
        installdir = "/install.386"
        if self.tdl.arch == "x86_64":
            installdir = "/install.amd"

        self.log.debug("Modifying isolinux.cfg")
        isolinuxcfg = os.path.join(self.iso_contents, "isolinux", "isolinux.cfg")
        extra = ""
        if self.config.need_auto_iso:
            extra = "auto=true "

        with open(isolinuxcfg, 'w') as f:
            f.write(textwrap.dedent(
                """\
                default customiso
                timeout 1
                prompt 0
                label customiso
                  menu label ^Customiso
                  menu default
                  kernel %s/vmlinuz
                  append file=/cdrom/preseed/customiso.seed %sdebian-installer/\
locale=en_US console-setup/layoutcode=us netcfg/choose_interface=auto priority=\
critical initrd=%s/initrd.gz --
                """ % (installdir, extra, installdir)))

    def get_auto_path(self):
        autoname = self.tdl.distro + self.tdl.update + ".auto"
        sp = self.tdl.update.split('.')
        if len(sp) == 3:
            autoname = self.tdl.distro + sp[0] + "." + sp[1] + ".auto"
        return oz.ozutil.generate_full_auto_path(autoname)

    def _generate_new_iso(self):
        """
        Method to create a new ISO based on the modified CD/DVD.
        """
        self.log.info("Generating new ISO")
        oz.ozutil.subprocess_check_output(["genisoimage", "-r", "-V", "Custom",
                                           "-J", "-l", "-no-emul-boot",
                                           "-b", "isolinux/isolinux.bin",
                                           "-c", "isolinux/boot.cat",
                                           "-boot-load-size", "4",
                                           "-cache-inodes", "-boot-info-table",
                                           "-v", "-o", self.output_iso,
                                           self.iso_contents],
                                          printfn=self.log.debug)

    def install(self, timeout=None, force=False):
        """
        Method to run the operating system installation.
        """
        extra = ""
        if self.config.need_auto_direct:
            extra = "auto=true "
        cmdline = "priority=critical " + extra + "locale=en_US"
        return self._do_install(timeout, force, 0, self.kernelfname,
                                self.initrdfname, cmdline)

    def _get_service_runlevel_link(self, g_handle, service):
        """
        Method to find the runlevel link(s) for a service based on the name
        and the (detected) default runlevel.
        """
        runlevel = self.get_default_runlevel(g_handle)

        lines = g_handle.cat('/etc/init.d/' + service).split("\n")
        startlevel = "99"
        for line in lines:
            if re.match('# chkconfig:', line):
                try:
                    startlevel = line.split(':')[1].split()[1]
                except Exception:
                    pass
                break

        return "/etc/rc" + runlevel + ".d/S" + startlevel + service

    def _image_ssh_teardown_step_1(self, g_handle):
        """
        First step to undo _image_ssh_setup (remove authorized keys).
        """
        self.log.debug("Teardown step 1")
        # reset the authorized keys
        self.log.debug("Resetting authorized_keys")
        g_handle.path_restore('/root/.ssh')

    def _image_ssh_teardown_step_2(self, g_handle):
        """
        Second step to undo _image_ssh_setup (reset sshd service).
        """
        self.log.debug("Teardown step 2")
        # remove custom sshd_config
        self.log.debug("Resetting sshd_config")
        g_handle.path_restore('/etc/ssh/sshd_config')

        # reset the service link
        self.log.debug("Resetting sshd service")
        if self.ssh_startuplink:
            g_handle.path_restore(self.ssh_startuplink)

    def _image_ssh_teardown_step_3(self, g_handle):
        """
        Fourth step to undo _image_ssh_setup (remove guest announcement).
        """
        self.log.debug("Teardown step 3")
        # remove announce cronjob
        self.log.debug("Resetting announcement to host")
        g_handle.remove_if_exists('/etc/cron.d/announce')

        # remove reportip
        self.log.debug("Removing reportip")
        g_handle.remove_if_exists('/root/reportip')

        # reset the service link
        self.log.debug("Resetting cron service")
        if self.cron_startuplink:
            g_handle.path_restore(self.cron_startuplink)

    def _image_ssh_teardown_step_4(self, g_handle):
        """
        Fourth step to undo changes by the operating system.  For instance,
        during first boot openssh generates ssh host keys and stores them
        in /etc/ssh.  Since this image might be cached later on, this method
        removes those keys.
        """
        for f in ["/etc/ssh/ssh_host_dsa_key", "/etc/ssh/ssh_host_dsa_key.pub",
                  "/etc/ssh/ssh_host_rsa_key", "/etc/ssh/ssh_host_rsa_key.pub",
                  "/etc/ssh/ssh_host_ecdsa_key", "/etc/ssh/ssh_host_ecdsa_key.pub",
                  "/etc/ssh/ssh_host_key", "/etc/ssh/ssh_host_key.pub"]:
            g_handle.remove_if_exists(f)

    def _image_ssh_setup_step_1(self, g_handle):
        """
        First step for allowing remote access (generate and upload ssh keys).
        """
        # part 1; upload the keys
        self.log.debug("Step 1: Uploading ssh keys")
        g_handle.path_backup('/root/.ssh')
        g_handle.mkdir('/root/.ssh')

        g_handle.path_backup('/root/.ssh/authorized_keys')

        self._generate_openssh_key(self.sshprivkey)

        g_handle.upload(self.sshprivkey + ".pub", '/root/.ssh/authorized_keys')

    def _image_ssh_setup_step_2(self, g_handle):
        """
        Second step for allowing remote access (configure sshd).
        """
        # part 2; check and setup sshd
        self.log.debug("Step 2: setup sshd")
        if not g_handle.exists('/usr/sbin/sshd'):
            raise oz.OzException.OzException("ssh not installed on the image, cannot continue")

        self.ssh_startuplink = self._get_service_runlevel_link(g_handle, 'ssh')
        g_handle.path_backup(self.ssh_startuplink)
        g_handle.ln_sf('/etc/init.d/ssh', self.ssh_startuplink)

        sshd_config_file = os.path.join(self.icicle_tmp, "sshd_config")
        f = open(sshd_config_file, 'w')
        f.write(self.sshd_config)
        f.close()

        try:
            g_handle.path_backup('/etc/ssh/sshd_config')
            g_handle.upload(sshd_config_file, '/etc/ssh/sshd_config')
        finally:
            os.unlink(sshd_config_file)

    def _image_ssh_setup_step_3(self, g_handle):
        """
        Third step for allowing remote access (make the guest announce itself
        on bootup).
        """
        # part 3; make sure the guest announces itself
        self.log.debug("Step 3: Guest announcement")
        if not g_handle.exists('/usr/sbin/cron'):
            raise oz.OzException.OzException("cron not installed on the image, cannot continue")

        scriptfile = os.path.join(self.icicle_tmp, "script")
        f = open(scriptfile, 'w')
        f.write("#!/bin/bash\n")
        f.write("/bin/sleep 20\n")
        f.write("DEV=$(/usr/bin/awk '{if ($2 == 0) print $1}' /proc/net/route) &&\n")
        f.write('[ -z "$DEV" ] && exit 0\n')
        f.write("ADDR=$(/sbin/ip -4 -o addr show dev $DEV | /usr/bin/awk '{print $4}' | /usr/bin/cut -d/ -f1) &&\n")
        f.write('[ -z "$ADDR" ] && exit 0\n')
        f.write('echo -n "!$ADDR,%s!" > /dev/ttyS1\n' % (self.uuid))
        f.close()
        try:
            g_handle.upload(scriptfile, '/root/reportip')
            g_handle.chmod(0o755, '/root/reportip')
        finally:
            os.unlink(scriptfile)

        announcefile = os.path.join(self.icicle_tmp, "announce")
        f = open(announcefile, 'w')
        f.write('*/1 * * * * root /bin/bash -c "/root/reportip"\n')
        f.close()
        try:
            g_handle.upload(announcefile, '/etc/cron.d/announce')
        finally:
            os.unlink(announcefile)

        self.cron_startuplink = self._get_service_runlevel_link(g_handle,
                                                                'cron')
        g_handle.path_backup(self.cron_startuplink)
        g_handle.ln_sf('/etc/init.d/cron', self.cron_startuplink)

    def _collect_setup(self, libvirt_xml):
        """
        Setup the guest for remote access.
        """
        self.log.info("Collection Setup")

        g_handle = oz.GuestFSManager.GuestFSLibvirtFactory(libvirt_xml, self.libvirt_conn)
        g_handle.mount_partitions()

        # we have to do 3 things to make sure we can ssh into Debian
        # 1)  Upload our ssh key
        # 2)  Make sure sshd is running on boot
        # 3)  Make the guest announce itself to the host

        try:
            try:
                self._image_ssh_setup_step_1(g_handle)

                try:
                    self._image_ssh_setup_step_2(g_handle)

                    try:
                        self._image_ssh_setup_step_3(g_handle)

                    except:
                        self._image_ssh_teardown_step_3(g_handle)
                        raise
                except:
                    self._image_ssh_teardown_step_2(g_handle)
                    raise
            except:
                self._image_ssh_teardown_step_1(g_handle)
                raise

        finally:
            g_handle.cleanup()

    def _collect_teardown(self, libvirt_xml):
        """
        Method to reverse the changes done in _collect_setup.
        """
        self.log.info("Collection Teardown")

        g_handle = oz.GuestFSManager.GuestFSLibvirtFactory(libvirt_xml, self.libvirt_conn)
        g_handle.mount_partitions()

        try:
            self._image_ssh_teardown_step_1(g_handle)

            self._image_ssh_teardown_step_2(g_handle)

            self._image_ssh_teardown_step_3(g_handle)

            self._image_ssh_teardown_step_4(g_handle)
        finally:
            g_handle.cleanup()
            shutil.rmtree(self.icicle_tmp)

    def _discover_repo_locality(self, repo_url, guestaddr, certdict):
        """
        Internal method to discover whether a repository is reachable from the
        guest or not.  It is used by customize_repos to decide which method to
        use to reach the repository.
        """

    # FIXME: Make this work for signed repositories
    def _customize_repos(self, guestaddr):
        """
        Method to generate and upload custom repository files based on the TDL.
        """

        self.log.debug("Installing additional repository files")

        for repo in list(self.tdl.repositories.values()):
            self.guest_execute_command(guestaddr, "echo '%s' > /etc/apt/sources.list.d/%s" % (repo.url.strip('\'"'), repo.name + ".list"))
            self.guest_execute_command(guestaddr, "apt-get update")

    def _install_packages(self, guestaddr, packstr):
        self.guest_execute_command(guestaddr,
                                   'apt-get install -y %s' % (packstr))

    def do_icicle(self, guestaddr):
        """
        Method to collect the package information and generate the ICICLE
        XML.
        """
        self.log.debug("Generating ICICLE")
        stdout, stderr_unused, retcode_unused = self.guest_execute_command(guestaddr,
                                                                           'dpkg --get-selections',
                                                                           timeout=30)

        # the data we get back from dpkg is in the form of:
        #
        # <package name>\t\t\tinstall
        #
        # so we have to strip out the tabs and the install before
        # passing it on to output_icicle_xml
        packages = []
        for line in stdout.split("\n"):
            packages.append(line.split("\t")[0])

        return self._output_icicle_xml(packages, self.tdl.description)

    def _get_kernel_from_txt_cfg(self, fetchurl):
        """
        Internal method to download and parse the txt.cfg file from a URL.  If
        the txt.cfg file does not exist, or it does not have the keys that we
        expect, this method raises an error.
        """
        txtcfgurl = fetchurl + "/debian-installer/" + self.debarch + "/boot-screens/txt.cfg"

        # first we check if the txt.cfg exists; this throws an exception if
        # it is missing
        info = oz.ozutil.http_get_header(txtcfgurl)
        if info['HTTP-Code'] != 200:
            raise oz.OzException.OzException("Could not find %s" % (txtcfgurl))

        txtcfg = os.path.join(self.icicle_tmp, "txt.cfg")
        self.log.debug("Going to write txt.cfg to %s", txtcfg)
        txtcfgfd = os.open(txtcfg, os.O_RDWR | os.O_CREAT | os.O_TRUNC)
        os.unlink(txtcfg)
        fp = os.fdopen(txtcfgfd)
        try:
            self.log.debug("Trying to get txt.cfg from " + txtcfgurl)
            oz.ozutil.http_download_file(txtcfgurl, txtcfgfd, False, self.log)

            # if we made it here, the txt.cfg existed.  Parse it and
            # find out the location of the kernel and ramdisk
            self.log.debug("Got txt.cfg, parsing")
            os.lseek(txtcfgfd, 0, os.SEEK_SET)
            # FIXME: This pattern doesn't match for Debian
            grub_pattern = re.compile(r"^default\s*(?P<default_entry>\w+)$.*"
                                      r"^label\s*(?P=default_entry)$.*"
                                      r"^\s*kernel\s*(?P<kernel>\S+)$.*"
                                      r"initrd=(?P<initrd>\S+).*"
                                      r"^label", re.DOTALL | re.MULTILINE)
            config_text = fp.read()
            match = re.search(grub_pattern, config_text)
            kernel = match.group('kernel')
            initrd = match.group('initrd')
        finally:
            fp.close()

        if kernel is None or initrd is None:
            raise oz.OzException.OzException("Empty kernel or initrd")

        self.log.debug("Returning kernel %s and initrd %s", kernel, initrd)
        return (kernel, initrd)

    def _gzip_file(self, inputfile, outputmode):
        """
        Internal method to gzip a file and write it to the initrd.
        """
        f = open(inputfile, 'rb')
        gzf = gzip.GzipFile(self.initrdfname, mode=outputmode)
        try:
            gzf.writelines(f)
            gzf.close()
            f.close()
        except:
            # there is a bit of asymmetry here in that OSs that support cpio
            # archives have the initial initrdfname copied in the higher level
            # function, but we delete it here.  OSs that don't support cpio,
            # though, get the initrd created right here.  C'est la vie
            os.unlink(self.initrdfname)
            raise

    def _create_cpio_initrd(self, preseedpath):
        """
        Internal method to create a modified CPIO initrd
        """
        extrafname = os.path.join(self.icicle_tmp, "extra.cpio")
        self.log.debug("Writing cpio to %s", extrafname)
        cpiofiledict = {}
        cpiofiledict[preseedpath] = 'preseed.cfg'
        oz.ozutil.write_cpio(cpiofiledict, extrafname)

        try:
            shutil.copyfile(self.initrdcache, self.initrdfname)
            self._gzip_file(extrafname, 'ab')
        finally:
            os.unlink(extrafname)

    def _initrd_inject_preseed(self, fetchurl, force_download):
        """
        Internal method to download and inject a preseed file into an initrd.
        """
        # we first see if we can use direct kernel booting, as that is
        # faster than downloading the ISO
        kernel = None
        initrd = None
        try:
            (kernel, initrd) = self._get_kernel_from_txt_cfg(fetchurl)
        except Exception:
            pass

        if kernel is None:
            self.log.debug("Kernel was None, trying debian-installer/%s/linux", self.debarch)
            # we couldn't find the kernel in the txt.cfg, so try a
            # hard-coded path
            kernel = "debian-installer/%s/linux" % (self.debarch)
        if initrd is None:
            self.log.debug("Initrd was None, trying debian-installer/%s/initrd.gz", self.debarch)
            # we couldn't find the initrd in the txt.cfg, so try a
            # hard-coded path
            initrd = "debian-installer/%s/initrd.gz" % (self.debarch)

        (fd, outdir) = oz.ozutil.open_locked_file(self.kernelcache)

        try:
            self._get_original_media('/'.join([self.url.rstrip('/'),
                                               kernel.lstrip('/')]),
                                     fd, outdir, force_download)

            # if we made it here, then we can copy the kernel into place
            shutil.copyfile(self.kernelcache, self.kernelfname)
        finally:
            os.close(fd)

        (fd, outdir) = oz.ozutil.open_locked_file(self.initrdcache)

        try:
            try:
                self._get_original_media('/'.join([self.url.rstrip('/'),
                                                   initrd.lstrip('/')]),
                                         fd, outdir, force_download)
            except:
                os.unlink(self.kernelfname)
                raise
        finally:
            os.close(fd)

        try:
            preseedpath = os.path.join(self.icicle_tmp, "preseed.cfg")
            self._copy_preseed(preseedpath)

            try:
                self._create_cpio_initrd(preseedpath)
            finally:
                os.unlink(preseedpath)
        except:
            os.unlink(self.kernelfname)
            raise

    def _remove_repos(self, guestaddr):
        # FIXME: until we switch over to doing repository add by hand (instead
        # of using add-apt-repository), we can't really reliably implement this
        pass

    def generate_install_media(self, force_download=False,
                               customize_or_icicle=False):
        """
        Method to generate the install media for Debian based operating
        systems.  If force_download is False (the default), then the
        original media will only be fetched if it is not cached locally.  If
        force_download is True, then the original media will be downloaded
        regardless of whether it is cached locally.
        """
        fetchurl = self.url
        if self.tdl.installtype == 'url':
            # set the fetchurl up-front so that if the OS doesn't support
            # initrd injection, or the injection fails for some reason, we
            # fall back to the mini.iso
            fetchurl += "/mini.iso"

            self.log.debug("Installtype is URL, trying to do direct kernel boot")
            try:
                return self._initrd_inject_preseed(self.url, force_download)
            except Exception as err:
                # if any of the above failed, we couldn't use the direct
                # kernel/initrd build method.  Fall back to trying to fetch
                # the mini.iso instead
                self.log.debug("Could not do direct boot, fetching mini.iso instead (the following error message is useful for bug reports, but can be ignored)")
                self.log.debug(err)

        return self._iso_generate_install_media(fetchurl, force_download,
                                                customize_or_icicle)

    def cleanup_install(self):
        """
        Method to cleanup any transient install data.
        """
        self.log.info("Cleaning up after install")

        for fname in [self.output_iso, self.initrdfname, self.kernelfname]:
            try:
                os.unlink(fname)
            except Exception:
                pass

        if not self.cache_original_media:
            for fname in [self.orig_iso, self.kernelcache, self.initrdcache]:
                try:
                    os.unlink(fname)
                except Exception:
                    pass


def get_class(tdl, config, auto, output_disk=None, netdev=None, diskbus=None,
              macaddress=None):
    """
    Factory method for Debian installs.
    """
    if tdl.update in version_to_config.keys():
        return DebianGuest(tdl, config, auto, output_disk, netdev, diskbus,
                           macaddress)


def get_supported_string():
    """
    Return supported versions as a string.
    """
    return "Debian: " + ", ".join(sorted(version_to_config.keys()))



# Copyright (C) 2010,2011  Chris Lalancette <clalance@redhat.com>
# Copyright (C) 2012-2017  Chris Lalancette <clalancette@gmail.com>

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

"""
RHEL-5 installation
"""

import os
import re

import oz.OzException
import oz.RedHat
import oz.ozutil


class RHEL5Configuration(object):
    """
    Configuration class for RHEL-5 installation.
    """
    def __init__(self, default_netdev, default_diskbus):
        self._default_netdev = default_netdev
        self._default_diskbus = default_diskbus

    @property
    def default_netdev(self):
        """
        Property method for the default netdev for this version of RHEL-5.
        """
        return self._default_netdev

    @property
    def default_diskbus(self):
        """
        Property method for the default diskbus for this version of RHEL-5.
        """
        return self._default_diskbus


version_to_config = {
    "U11": RHEL5Configuration(default_netdev='virtio', default_diskbus='virtio'),
    "U10": RHEL5Configuration(default_netdev='virtio', default_diskbus='virtio'),
    "U9": RHEL5Configuration(default_netdev='virtio', default_diskbus='virtio'),
    "U8": RHEL5Configuration(default_netdev='virtio', default_diskbus='virtio'),
    "U7": RHEL5Configuration(default_netdev='virtio', default_diskbus='virtio'),
    "U6": RHEL5Configuration(default_netdev='virtio', default_diskbus='virtio'),
    "U5": RHEL5Configuration(default_netdev='virtio', default_diskbus='virtio'),
    "U4": RHEL5Configuration(default_netdev='virtio', default_diskbus='virtio'),
    "U3": RHEL5Configuration(default_netdev=None, default_diskbus=None),
    "U2": RHEL5Configuration(default_netdev=None, default_diskbus=None),
    "U1": RHEL5Configuration(default_netdev=None, default_diskbus=None),
    "GOLD": RHEL5Configuration(default_netdev=None, default_diskbus=None),
}


class RHEL5Guest(oz.RedHat.RedHatLinuxCDYumGuest):
    """
    Class for RHEL-5 GOLD, U1, U2, U3, U4, U5, U6, U7, U8, U9, U10 and U11 installation.
    """
    def __init__(self, tdl, config, auto, nicmodel, diskbus, output_disk=None,
                 macaddress=None):
        self.config = version_to_config[tdl.update]
        if nicmodel is None:
            nicmodel = self.config.default_netdev
        if diskbus is None:
            diskbus = self.config.default_diskbus
        oz.RedHat.RedHatLinuxCDYumGuest.__init__(self, tdl, config, auto,
                                                 output_disk, nicmodel, diskbus,
                                                 True, True, "cpio", macaddress,
                                                 True)

    def _modify_iso(self):
        """
        Method to make the boot ISO auto-boot with appropriate parameters.
        """
        self._copy_kickstart(os.path.join(self.iso_contents, "ks.cfg"))

        initrdline = "  append initrd=initrd.img ks=cdrom:/ks.cfg method="
        if self.tdl.installtype == "url":
            initrdline += self.url
        else:
            initrdline += "cdrom:/dev/cdrom"
        self._modify_isolinux(initrdline)

    def _check_pvd(self):
        """
        Method to ensure that boot ISO is a DVD (we cannot use boot CDs to
        install RHEL-5/CentOS-5 since it requires a switch during install,
        which we cannot detect).
        """
        with open(self.orig_iso, "r") as cdfd:
            pvd = self._get_primary_volume_descriptor(cdfd)

        # all of the below should have "LINUX" as their system_identifier,
        # so check it here
        if pvd.system_identifier != "LINUX                           ":
            raise oz.OzException.OzException("Invalid system identifier on ISO for " + self.tdl.distro + " install")

        if self.tdl.distro == "RHEL-5":
            if self.tdl.installtype == 'iso':
                if not re.match(r"RHEL/5(\.[0-9]{1,2})? " + self.tdl.arch + " DVD",
                                pvd.volume_identifier):
                    raise oz.OzException.OzException("Only DVDs are supported for RHEL-5 ISO installs")
            else:
                # url installs
                if not pvd.volume_identifier.startswith("Red Hat Enterprise Linux"):
                    raise oz.OzException.OzException("Invalid boot.iso for RHEL-5 URL install")
        elif self.tdl.distro == "CentOS-5":
            # CentOS-5
            if self.tdl.installtype == 'iso':
                # unfortunately CentOS-5 has the same volume identifier for both
                # DVDs and CDs.  To tell them apart, we assume that if the
                # size is smaller than 1GB, this is a CD
                if not re.match(r"CentOS_5.[0-9]{1,2}_Final", pvd.volume_identifier) or (pvd.space_size * 2048) < 1 * 1024 * 1024 * 1024:
                    raise oz.OzException.OzException("Only DVDs are supported for CentOS-5 ISO installs")
            else:
                # url installs
                if not re.match(r"CentOS *", pvd.volume_identifier):
                    raise oz.OzException.OzException("Invalid boot.iso for CentOS-5 URL install")
        elif self.tdl.distro == "SLC-5":
            # SLC-5
            if self.tdl.installtype == 'iso':
                if not re.match(r"Scientific Linux CERN 5.[0-9]{1,2}", pvd.volume_identifier):
                    raise oz.OzException.OzException("Only DVDs are supported for SLC-5 ISO installs")
            else:
                # url installs
                if not re.match(r"CentOS *", pvd.volume_identifier):
                    raise oz.OzException.OzException("Invalid boot.iso for SLC-5 URL install")

    def get_auto_path(self):
        """
        Method to create the correct path to the RHEL 5 kickstart file.
        """
        return oz.ozutil.generate_full_auto_path("RHEL5.auto")


def get_class(tdl, config, auto, output_disk=None, netdev=None, diskbus=None,
              macaddress=None):
    """
    Factory method for RHEL-5 installs.
    """
    if tdl.update in version_to_config.keys():
        return RHEL5Guest(tdl, config, auto, netdev, diskbus, output_disk,
                          macaddress)


def get_supported_string():
    """
    Return supported versions as a string.
    """
    return "RHEL/OL/CentOS/Scientific Linux{,CERN} 5: " + ", ".join(sorted(version_to_config.keys()))



# Copyright (C) 2010,2011  Chris Lalancette <clalance@redhat.com>
# Copyright (C) 2012-2017  Chris Lalancette <clalancette@gmail.com>

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

"""
RHEL-3 installation
"""

import os
import re

import oz.OzException
import oz.RedHat
import oz.ozutil

versions = ["GOLD", "U1", "U2", "U3", "U4", "U5", "U6", "U7", "U8", "U9"]


class RHEL3Guest(oz.RedHat.RedHatLinuxCDGuest):
    """
    Class for RHEL-3 GOLD, U1, U2, U3, U4, U5, U6, U7, U8, and U9 installation.
    """
    def __init__(self, tdl, config, auto, output_disk, netdev, diskbus,
                 macaddress):
        iso_support = True
        if tdl.distro == "RHEL-3":
            iso_support = False

        # although we could use ext2 for the initrdtype here (and hence get
        # fast initial installs), it isn't super reliable on RHEL-3.  Just
        # disable it and fall back to the boot.iso method which is more reliable
        oz.RedHat.RedHatLinuxCDGuest.__init__(self, tdl, config, auto,
                                              output_disk, netdev, diskbus,
                                              iso_support, True, None,
                                              macaddress)

        # override the sshd_config value set in RedHatLinuxCDGuest.__init__
        self.sshd_config = """\
SyslogFacility AUTHPRIV
PasswordAuthentication yes
ChallengeResponseAuthentication no
X11Forwarding yes
Subsystem	sftp	/usr/libexec/openssh/sftp-server
"""

    def _modify_iso(self):
        """
        Method to make the boot ISO auto-boot with appropriate parameters.
        """
        self._copy_kickstart(os.path.join(self.iso_contents, "ks.cfg"))

        initrdline = "  append initrd=initrd.img ks=cdrom:/ks.cfg method="
        if self.tdl.installtype == "url":
            initrdline += self.url
        else:
            initrdline += "cdrom:/dev/cdrom"
        self._modify_isolinux(initrdline)

    def _check_pvd(self):
        """
        Method to ensure the the boot ISO for an ISO install is a DVD
        """
        cdfd = open(self.orig_iso, "r")
        pvd = self._get_primary_volume_descriptor(cdfd)
        cdfd.close()

        if pvd.system_identifier != "LINUX                           ":
            raise oz.OzException.OzException("Invalid system identifier on ISO for " + self.tdl.distro + " install")

        if self.tdl.distro == "RHEL-3":
            if self.tdl.installtype == "iso":
                raise oz.OzException.OzException("BUG: shouldn't be able to reach RHEL-3 with ISO checking")
            # The boot ISOs for RHEL-3 don't have a whole lot of identifying
            # information.  We just pass through here, doing nothing
        else:
            if self.tdl.installtype == "iso":
                centos_disk_1 = re.match(r"CentOS-3(\.[0-9])? Disk 1", pvd.volume_identifier)
                centos_server = re.match(r"CentOS-3(\.[0-9])? server", pvd.volume_identifier)
                centos_dvd = re.match(r"CentOS-3(\.[0-9])? " + self.tdl.arch + " DVD", pvd.volume_identifier)
                if not centos_disk_1 and not centos_server and not centos_dvd:
                    raise oz.OzException.OzException("Only DVDs are supported for CentOS-3 ISO installs")
            # The boot ISOs for CentOS-3 don't have a whole lot of identifying
            # information.  We just pass through here, doing nothing

    def get_auto_path(self):
        """
        Method to create the correct path to the RHEL 3 kickstart files.
        """
        return oz.ozutil.generate_full_auto_path("RHEL3.auto")


def get_class(tdl, config, auto, output_disk=None, netdev=None, diskbus=None,
              macaddress=None):
    """
    Factory method for RHEL-3 installs.
    """
    if tdl.update in versions:
        return RHEL3Guest(tdl, config, auto, output_disk, netdev, diskbus,
                          macaddress)


def get_supported_string():
    """
    Return supported versions as a string.
    """
    return "RHEL/CentOS 3: " + ", ".join(versions)



# Copyright (C) 2013  harmw <harm@weites.com>
# Copyright (C) 2013-2017  Chris Lalancette <clalancette@gmail.com>

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

"""
FreeBSD installation
"""

import os

import oz.Guest
import oz.OzException
import oz.ozutil


class FreeBSDConfiguration(object):
    """
    Configuration class for FreeBSD.
    """
    def __init__(self, default_netdev, default_diskbus):
        self._default_netdev = default_netdev
        self._default_diskbus = default_diskbus

    @property
    def default_netdev(self):
        """
        Property method for the default netdev for this FreeBSD version.
        """
        return self._default_netdev

    @property
    def default_diskbus(self):
        """
        Property method for the default diskbus for this FreeBSD version.
        """
        return self._default_diskbus


version_to_config = {
    "11.0": FreeBSDConfiguration(default_netdev='virtio', default_diskbus='virtio'),
    "10.3": FreeBSDConfiguration(default_netdev='virtio', default_diskbus='virtio'),
    "10.2": FreeBSDConfiguration(default_netdev='virtio', default_diskbus='virtio'),
    "10.1": FreeBSDConfiguration(default_netdev='virtio', default_diskbus='virtio'),
    "10.0": FreeBSDConfiguration(default_netdev='virtio', default_diskbus='virtio'),
}


class FreeBSD(oz.Guest.CDGuest):
    """
    Class for FreeBSD 10.0, 10.1, 10.2, 10.3 and 11.0 installation.
    """
    def __init__(self, tdl, config, auto, output_disk, netdev, diskbus,
                 macaddress):
        self.config = version_to_config[tdl.update]
        if netdev is None:
            netdev = self.config.default_netdev
        if diskbus is None:
            diskbus = self.config.default_diskbus
        oz.Guest.CDGuest.__init__(self, tdl, config, auto, output_disk,
                                  netdev, "localtime", "usb", diskbus, True,
                                  False, macaddress)

    def _generate_new_iso(self):
        """
        Method to create a new ISO based on the modified CD/DVD.
        """
        self.log.debug("Generating new ISO")
        oz.ozutil.subprocess_check_output(["genisoimage",
                                           "-R", "-no-emul-boot",
                                           "-b", "boot/cdboot", "-v",
                                           "-o", self.output_iso,
                                           self.iso_contents],
                                          printfn=self.log.debug)

    def _modify_iso(self):
        """
        Method to make the boot ISO auto-boot with appropriate parameters.
        """
        self.log.debug("Modifying ISO")

        def _replace(line):
            """
            Method that is called back from copy_modify_file to replace
            the rootpassword in the installerconfig file
            """
            keys = {
                '#ROOTPW#': self.rootpw,
            }
            for key in keys:
                line = line.replace(key, keys[key])
            return line

        # Copy the installconfig file to /etc/ on the iso image so bsdinstall(8)
        # can use that to do an unattended installation. This rules file
        # contains both setup rules and a post script. This stage also prepends
        # the post script with additional commands so it's possible to install
        # extra	packages specified in the .tdl file.

        outname = os.path.join(self.iso_contents, "etc", "installerconfig")
        oz.ozutil.copy_modify_file(self.auto, outname, _replace)

        # Make sure the iso can be mounted at boot, otherwise this error shows
        # up after booting the kernel:
        #  mountroot: waiting for device /dev/iso9660/FREEBSD_INSTALL ...
        #  Mounting from cd9660:/dev/iso9660/FREEBSD_INSTALL failed with error 19.

        loaderconf = os.path.join(self.iso_contents, "boot", "loader.conf")
        with open(loaderconf, 'w') as conf:
            conf.write('autoboot_delay="0"\n')
            conf.write('vfs.root.mountfrom="cd9660:/dev/cd0"\n')
            conf.write('console="vidconsole comconsole"\n')
            conf.write('kern.panic_reboot_wait_time="-1"\n')


def get_class(tdl, config, auto, output_disk=None, netdev=None, diskbus=None,
              macaddress=None):
    """
    Factory method for FreeBSD installs.
    """
    if tdl.update in version_to_config.keys():
        return FreeBSD(tdl, config, auto, output_disk, netdev, diskbus,
                       macaddress)


def get_supported_string():
    """
    Return supported versions as a string.
    """
    return "FreeBSD: " + ", ".join(sorted(version_to_config.keys()))



# Copyright (C) 2013-2018  Chris Lalancette <clalancette@gmail.com>

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

"""
Mageia installation
"""

import os
import re
import shutil
try:
    import urllib.parse as urlparse
except ImportError:
    import urlparse

import oz.GuestFSManager
import oz.Linux
import oz.OzException
import oz.ozutil


class MageiaConfiguration(object):
    """
    Configuration class for Mageia installation.
    """
    def __init__(self, isolinux_style, default_netdev, default_diskbus):
        self._isolinux_style = isolinux_style
        self._default_netdev = default_netdev
        self._default_diskbus = default_diskbus

    @property
    def isolinux_style(self):
        """
        Property method for the 'old' or 'new' isolinux style.
        """
        return self._isolinux_style

    @property
    def default_netdev(self):
        """
        Property method for the default netdev for this version of Mageia.
        """
        return self._default_netdev

    @property
    def default_diskbus(self):
        """
        Property method for the default diskbus for this version of Mageia.
        """
        return self._default_diskbus


version_to_config = {
    '5': MageiaConfiguration(isolinux_style="new", default_netdev='virtio',
                             default_diskbus='virtio'),
    '4.1': MageiaConfiguration(isolinux_style="new", default_netdev='virtio',
                               default_diskbus='virtio'),
    '4': MageiaConfiguration(isolinux_style="new", default_netdev='virtio',
                             default_diskbus='virtio'),
    '3': MageiaConfiguration(isolinux_style="old", default_netdev='virtio',
                             default_diskbus='virtio'),
    '2': MageiaConfiguration(isolinux_style="old", default_netdev='virtio',
                             default_diskbus='virtio'),
}


class MageiaGuest(oz.Linux.LinuxCDGuest):
    """
    Class for Mageia 2, 3, 4, 4.1, and 5 installation.
    """
    def __init__(self, tdl, config, auto, output_disk, netdev, diskbus,
                 macaddress):
        self.config = version_to_config[tdl.update]
        if netdev is None:
            netdev = self.config.default_netdev
        if diskbus is None:
            diskbus = self.config.default_diskbus
        oz.Linux.LinuxCDGuest.__init__(self, tdl, config, auto, output_disk,
                                       netdev, diskbus, True, True,
                                       macaddress)

        self.sshd_was_active = False
        self.crond_was_active = False

    def _modify_iso(self):
        """
        Method to make the boot ISO auto-boot with appropriate parameters.
        """
        self.log.debug("Modifying ISO")

        self.log.debug("Copying cfg file to floppy image")

        outname = os.path.join(self.iso_contents, "auto_inst.cfg")

        if self.default_auto_file():

            def _cfg_sub(line):
                """
                Method that is called back from oz.ozutil.copy_modify_file() to
                modify preseed files as appropriate for Mageia.
                """
                if re.search("'password' =>", line):
                    return "			'password' => '" + self.rootpw + "',\n"
                return line

            oz.ozutil.copy_modify_file(self.auto, outname, _cfg_sub)
        else:
            shutil.copy(self.auto, outname)

        oz.ozutil.subprocess_check_output(["/sbin/mkfs.msdos", "-C",
                                           self.output_floppy, "1440"])
        oz.ozutil.subprocess_check_output(["mcopy", "-n", "-o", "-i",
                                           self.output_floppy, outname,
                                           "::AUTO_INST.CFG"])

        self.log.debug("Modifying isolinux.cfg")
        if self.config.isolinux_style == "old":
            # Mageia 2 dual   - isolinux/32.cfg
            #                   isolinux/64.cfg
            #                   isolinux/alt0/32/vmlinuz
            #                   isolinux/alt0/32/all.rdz
            #                   isolinux/alt0/64/vmlinuz
            #                   isolinux/alt0/64/all.rdz
            # Mageia 2 x86_64 - x86_64/isolinux/isolinux.cfg
            #                   x86_64/isolinux/alt0/vmlinuz
            #                   x86_64/isolinux/alt0/all.rdz
            # Mageia 2 i586   - i586/isolinux/isolinux.cfg
            #                   i586/isolinux/vmlinuz
            #                   i586/isolinux/all.rdz
            # Mageia 3 dual   - syslinux/32.cfg
            #                   syslinux/64.cfg
            #                   syslinux/alt0/32/vmlinuz
            #                   syslinux/alt0/32/all.rdz
            #                   syslinux/alt0/64/vmlinuz
            #                   syslinux/alt0/64/all.rdz
            # Mageia 3 x86_64 - x86_64/isolinux/isolinux.cfg
            #                   x86_64/isolinux/alt0/vmlinuz
            #                   x86_64/isolinux/alt0/all.rdz
            # Mageia 3 i586   - i586/isolinux/isolinux.cfg
            #                   i586/isolinux/alt0/vmlinuz
            #                   i586/isolinux/alt0/all.rdz
            isolinuxstr = None
            if os.path.exists(os.path.join(self.iso_contents, 'isolinux')):
                isolinuxstr = "isolinux"
            elif os.path.exists(os.path.join(self.iso_contents, 'syslinux')):
                isolinuxstr = "syslinux"

            if isolinuxstr is not None:
                if self.tdl.arch == "i386":
                    mageia_arch = "32"
                else:
                    mageia_arch = "64"

                # This looks like a dual CD, so let's set things up that way.
                isolinuxcfg = os.path.join(self.iso_contents, isolinuxstr, mageia_arch + ".cfg")
                self.isolinuxbin = os.path.join(isolinuxstr, mageia_arch + ".bin")
                kernel = "alt0/" + mageia_arch + "/vmlinuz"
                initrd = "alt0/" + mageia_arch + "/all.rdz"
            else:
                # This looks like an i586 or x86_64 ISO, so set things up that way.
                mageia_arch = self.tdl.arch
                if self.tdl.arch == "i386":
                    mageia_arch = "i586"
                isolinuxcfg = os.path.join(self.iso_contents, mageia_arch, 'isolinux', 'isolinux.cfg')
                self.isolinuxbin = os.path.join(mageia_arch, 'isolinux', 'isolinux.bin')
                kernel = "alt0/vmlinuz"
                initrd = "alt0/all.rdz"
            flags = "ramdisk_size=128000 root=/dev/ram3 acpi=ht vga=788 automatic=method:cdrom"
        else:
            # Mageia 4 dual     - isolinux/i586.cfg
            #                     isolinux/x86_64.cfg
            #                     isolinux/i586/vmlinuz
            #                     isolinux/i586/all.rdz
            #                     isolinux/x86_64/vmlinuz
            #                     isolinux/x86_64/all.rdz
            # Mageia 4 x86_64   - isolinux/isolinux.cfg
            #                     isolinux/x86_64/vmlinuz
            #                     isolinux/x86_64/all.rdz
            # Mageia 4 i586     - isolinux/isolinux.cfg
            #                     isolinux/i586/vmlinuz
            #                     isolinuz/i586/all.rdz
            # Mageia 4.1 dual   - isolinux/i586.cfg
            #                     isolinux/x86_64.cfg
            #                     isolinux/i586/vmlinuz
            #                     isolinux/i586/all.rdz
            #                     isolinux/x86_64/vmlinuz
            #                     isolinux/x86_64/all.rdz
            # Mageia 4.1 x86_64 - isolinux/isolinux.cfg
            #                     isolinux/x86_64/vmlinuz
            #                     isolinux/x86_64/all.rdz
            # Mageia 4.1 i586 -   isolinux/isolinux.cfg
            #                     isolinux/i586/vmlinuz
            #                     isolinux/i586/all.rdz
            # Mageia 5 dual     - isolinux/i586.cfg
            #                     isolinux/x86_64.cfg
            #                     isolinux/i586/vmlinuz
            #                     isolinux/i586/all.rdz
            #                     isolinux/x86_64/vmlinuz
            #                     isolinux/x86_64/all.rdz
            # Mageia 5 x86_64   - isolinux/isolinux.cfg
            #                     isolinux/x86_64/vmlinuz
            #                     isolinux/x86_64/all.rdz
            # Mageia 5 i586 -     isolinux/isolinux.cfg
            #                     isolinux/i586/vmlinuz
            #                     isolinux/i586/all.rdz

            # Starting with Mageia 4, things are a lot more regular.  The
            # directory always starts with isolinux.  If it is a dual ISO, then
            # there is an i586.cfg and x86_64.cfg describing how to boot each
            # of them.  Otherwise, there is just an isolinux.cfg.  The kernel
            # and initrd are always in the same place.
            mageia_arch = self.tdl.arch
            if self.tdl.arch == "i386":
                mageia_arch = "i586"
            if os.path.exists(os.path.join(self.iso_contents, 'isolinux', 'i586.cfg')):
                # A dual, so use the correct cfg
                isolinuxcfg = os.path.join(self.iso_contents, 'isolinux', mageia_arch + ".cfg")
                self.isolinuxbin = os.path.join('isolinux', mageia_arch + ".bin")
            else:
                isolinuxcfg = os.path.join(self.iso_contents, 'isolinux', 'isolinux.cfg')
                self.isolinuxbin = os.path.join('isolinux', 'isolinux.bin')
            kernel = mageia_arch + "/vmlinuz"
            initrd = mageia_arch + "/all.rdz"
            if self.tdl.installtype == "url":
                url = urlparse.urlparse(self.tdl.url)
                flags = "automatic=method:%s,ser:%s,dir:%s,int:eth0,netw:dhcp" % (url.scheme, url.hostname, url.path)
            else:
                flags = "automatic=method:cdrom"

        with open(isolinuxcfg, 'w') as f:
            f.write("""\
default customiso
timeout 1
prompt 0
label customiso
  kernel %s
  append initrd=%s kickstart=floppy %s
""" % (kernel, initrd, flags))

    def _generate_new_iso(self):
        """
        Method to create a new ISO based on the modified CD/DVD.
        """
        self.log.info("Generating new ISO")

        oz.ozutil.subprocess_check_output(["genisoimage", "-r", "-V", "Custom",
                                           "-J", "-l", "-no-emul-boot",
                                           "-b", self.isolinuxbin,
                                           "-c", "boot.catalog",
                                           "-boot-load-size", "4",
                                           "-cache-inodes", "-boot-info-table",
                                           "-v", "-o", self.output_iso,
                                           self.iso_contents],
                                          printfn=self.log.debug)

    def _get_service_runlevel_link(self, g_handle, service):
        """
        Method to find the runlevel link(s) for a service based on the name
        and the (detected) default runlevel.
        """
        runlevel = self.get_default_runlevel(g_handle)

        lines = g_handle.cat('/etc/init.d/' + service).split("\n")
        startlevel = "99"
        for line in lines:
            if re.match('# chkconfig:', line):
                try:
                    startlevel = line.split(':')[1].split()[1]
                except Exception:
                    pass
                break

        return "/etc/rc" + runlevel + ".d/S" + startlevel + service

    def _image_ssh_teardown_step_1(self, g_handle):
        """
        First step to undo _image_ssh_setup (remove authorized keys).
        """
        self.log.debug("Teardown step 1")
        # reset the authorized keys
        self.log.debug("Resetting authorized_keys")
        g_handle.path_restore('/root/.ssh/authorized_keys')

    def _image_ssh_teardown_step_2(self, g_handle):
        """
        Second step to undo _image_ssh_setup (remove custom sshd_config).
        """
        self.log.debug("Teardown step 2")
        # remove custom sshd_config
        self.log.debug("Resetting sshd_config")
        g_handle.path_restore('/etc/ssh/sshd_config')

        # reset the service link
        self.log.debug("Resetting sshd service")
        if g_handle.exists('/usr/lib/systemd/system/sshd.service'):
            if not self.sshd_was_active:
                g_handle.rm('/etc/systemd/system/multi-user.target.wants/sshd.service')
        else:
            startuplink = self._get_service_runlevel_link(g_handle, 'sshd')
            g_handle.path_restore(startuplink)

    def _image_ssh_teardown_step_3(self, g_handle):
        """
        Third step to undo _image_ssh_setup (remove guest announcement).
        """
        self.log.debug("Teardown step 3")
        # remove announce cronjob
        self.log.debug("Resetting announcement to host")
        g_handle.remove_if_exists('/etc/cron.d/announce')

        # remove reportip
        self.log.debug("Removing reportip")
        g_handle.remove_if_exists('/root/reportip')

        # reset the service link
        self.log.debug("Resetting cron service")
        if g_handle.exists('/usr/lib/systemd/system/cron.service'):
            if not self.crond_was_active:
                g_handle.rm('/etc/systemd/system/multi-user.target.wants/cron.service')
        else:
            runlevel = self.get_default_runlevel(g_handle)
            startuplink = '/etc/rc.d/rc' + runlevel + ".d/S06cron"
            g_handle.path_restore(startuplink)

    def _image_ssh_teardown_step_4(self, g_handle):
        """
        Fourth step to undo changes by the operating system.  For instance,
        during first boot openssh generates ssh host keys and stores them
        in /etc/ssh.  Since this image might be cached later on, this method
        removes those keys.
        """
        for f in ["/etc/ssh/ssh_host_dsa_key", "/etc/ssh/ssh_host_dsa_key.pub",
                  "/etc/ssh/ssh_host_rsa_key", "/etc/ssh/ssh_host_rsa_key.pub",
                  "/etc/ssh/ssh_host_ecdsa_key", "/etc/ssh/ssh_host_ecdsa_key.pub",
                  "/etc/ssh/ssh_host_key", "/etc/ssh/ssh_host_key.pub"]:
            g_handle.remove_if_exists(f)

        # Remove any lease files; this is so that subsequent boots don't try
        # to connect to a DHCP server that is on a totally different network
        for lease in g_handle.glob_expand("/var/lib/dhcp/*.leases"):
            g_handle.rm(lease)

        for lease in g_handle.glob_expand("/var/lib/dhcp6/*.leases"):
            g_handle.rm(lease)

        for lease in g_handle.glob_expand("/var/lib/dhcp6/*.lease"):
            g_handle.rm(lease)

    def _collect_teardown(self, libvirt_xml):
        """
        Method to reverse the changes done in _collect_setup.
        """
        self.log.info("Collection Teardown")

        g_handle = oz.GuestFSManager.GuestFSLibvirtFactory(libvirt_xml, self.libvirt_conn)
        g_handle.mount_partitions()

        try:
            self._image_ssh_teardown_step_1(g_handle)

            self._image_ssh_teardown_step_2(g_handle)

            self._image_ssh_teardown_step_3(g_handle)

            self._image_ssh_teardown_step_4(g_handle)
        finally:
            g_handle.cleanup()
            shutil.rmtree(self.icicle_tmp)

    def _image_ssh_setup_step_1(self, g_handle):
        """
        First step for allowing remote access (generate and upload ssh keys).
        """
        # part 1; upload the keys
        self.log.debug("Step 1: Uploading ssh keys")
        if not g_handle.exists('/root/.ssh'):
            g_handle.mkdir('/root/.ssh')

        g_handle.path_backup('/root/.ssh/authorized_keys')

        self._generate_openssh_key(self.sshprivkey)

        g_handle.upload(self.sshprivkey + ".pub", '/root/.ssh/authorized_keys')

    def _image_ssh_setup_step_2(self, g_handle):
        """
        Second step for allowing remote access (ensure sshd is running).
        """
        # part 2; check and setup sshd
        self.log.debug("Step 2: setup sshd")
        if not g_handle.exists('/usr/sbin/sshd'):
            raise oz.OzException.OzException("ssh not installed on the image, cannot continue")

        if g_handle.exists('/usr/lib/systemd/system/sshd.service'):
            if g_handle.exists('/etc/systemd/system/multi-user.target.wants/sshd.service'):
                self.sshd_was_active = True
            else:
                g_handle.ln_sf('/usr/lib/systemd/system/sshd.service',
                               '/etc/systemd/system/multi-user.target.wants/sshd.service')
        else:
            startuplink = self._get_service_runlevel_link(g_handle, 'sshd')
            g_handle.path_backup(startuplink)
            g_handle.ln_sf('/etc/init.d/sshd', startuplink)

        sshd_config_file = self.icicle_tmp + "/sshd_config"
        with open(sshd_config_file, 'w') as f:
            f.write("""PasswordAuthentication no
UsePAM yes

X11Forwarding yes

Subsystem      sftp    /usr/lib64/ssh/sftp-server

AcceptEnv LANG LC_CTYPE LC_NUMERIC LC_TIME LC_COLLATE LC_MONETARY LC_MESSAGES
AcceptEnv LC_PAPER LC_NAME LC_ADDRESS LC_TELEPHONE LC_MEASUREMENT
AcceptEnv LC_IDENTIFICATION LC_ALL
""")

        try:
            g_handle.path_backup("/etc/ssh/sshd_config")
            g_handle.upload(sshd_config_file, '/etc/ssh/sshd_config')
        finally:
            os.unlink(sshd_config_file)

    def _image_ssh_setup_step_3(self, g_handle):
        """
        Third step for allowing remote access (make the guest announce itself
        on bootup).
        """
        # part 3; make sure the guest announces itself
        self.log.debug("Step 3: Guest announcement")

        scriptfile = os.path.join(self.icicle_tmp, "script")

        if not g_handle.exists('/usr/sbin/crond'):
            raise oz.OzException.OzException("cron not installed on the image, cannot continue")

        with open(scriptfile, 'w') as f:
            f.write("""\
#!/bin/bash
DEV=$(/bin/awk '{if ($2 == 0) print $1}' /proc/net/route) &&
[ -z "$DEV" ] && exit 0
ADDR=$(/sbin/ip -4 -o addr show dev $DEV | /bin/awk '{print $4}' | /usr/bin/cut -d/ -f1) &&
[ -z "$ADDR" ] && exit 0
echo -n "!$ADDR,%s!" > /dev/ttyS1
""" % (self.uuid))

        try:
            g_handle.upload(scriptfile, '/root/reportip')
            g_handle.chmod(0o755, '/root/reportip')
        finally:
            os.unlink(scriptfile)

        announcefile = os.path.join(self.icicle_tmp, "announce")
        with open(announcefile, 'w') as f:
            f.write('*/1 * * * * root /bin/bash -c "/root/reportip"\n')

        try:
            g_handle.upload(announcefile, '/etc/cron.d/announce')
        finally:
            os.unlink(announcefile)

        if g_handle.exists('/usr/lib/systemd/system/cron.service'):
            if g_handle.exists('/etc/systemd/system/multi-user.target.wants/cron.service'):
                self.crond_was_active = True
            else:
                g_handle.ln_sf('/lib/systemd/system/cron.service',
                               '/etc/systemd/system/multi-user.target.wants/cron.service')
        else:
            runlevel = self.get_default_runlevel(g_handle)
            startuplink = '/etc/rc.d/rc' + runlevel + ".d/S06cron"
            g_handle.path_backup(startuplink)
            g_handle.ln_sf('/etc/init.d/cron', startuplink)

    def _collect_setup(self, libvirt_xml):
        """
        Setup the guest for remote access.
        """
        self.log.info("Collection Setup")

        g_handle = oz.GuestFSManager.GuestFSLibvirtFactory(libvirt_xml, self.libvirt_conn)
        g_handle.mount_partitions()

        # we have to do 3 things to make sure we can ssh into OpenSUSE:
        # 1)  Upload our ssh key
        # 2)  Make sure sshd is running on boot
        # 3)  Make the guest announce itself to the host

        try:
            try:
                self._image_ssh_setup_step_1(g_handle)

                try:
                    self._image_ssh_setup_step_2(g_handle)

                    try:
                        self._image_ssh_setup_step_3(g_handle)
                    except:
                        self._image_ssh_teardown_step_3(g_handle)
                        raise
                except:
                    self._image_ssh_teardown_step_2(g_handle)
                    raise
            except:
                self._image_ssh_teardown_step_1(g_handle)
                raise

        finally:
            g_handle.cleanup()

    def do_icicle(self, guestaddr):
        """
        Method to collect the package information and generate the ICICLE
        XML.
        """
        self.log.debug("Generating ICICLE")
        stdout, stderr_unused, retcode_unused = self.guest_execute_command(guestaddr,
                                                                           'rpm -qa',
                                                                           timeout=30)

        package_split = stdout.split("\n")

        extrasplit = None
        if self.tdl.icicle_extra_cmd:
            extrastdout, stderr_unused, retcode_unused = self.guest_execute_command(guestaddr,
                                                                                    self.tdl.icicle_extra_cmd,
                                                                                    timeout=30)
            extrasplit = extrastdout.split("\n")

            if len(package_split) != len(extrasplit):
                raise oz.OzException.OzException("Invalid extra package command; it must return the same set of packages as 'rpm -qa'")

        return self._output_icicle_xml(package_split, self.tdl.description,
                                       extrasplit)

    def generate_install_media(self, force_download=False,
                               customize_or_icicle=False):
        """
        Method to generate the install media for Mageia based operating
        systems.  If force_download is False (the default), then the
        original media will only be fetched if it is not cached locally.  If
        force_download is True, then the original media will be downloaded
        regardless of whether it is cached locally.
        """
        fetchurl = self.url
        if self.tdl.installtype == 'url':
            fetchurl += "/install/images/boot.iso"
        return self._iso_generate_install_media(fetchurl, force_download,
                                                customize_or_icicle)

    def install(self, timeout=None, force=False):
        fddev = self._InstallDev("floppy", self.output_floppy, "fda")
        return self._do_install(timeout, force, 0, None, None, None,
                                [fddev])


def get_class(tdl, config, auto, output_disk=None, netdev=None, diskbus=None,
              macaddress=None):
    """
    Factory method for Mageia installs.
    """
    if tdl.update in version_to_config.keys():
        return MageiaGuest(tdl, config, auto, output_disk, netdev, diskbus,
                           macaddress)


def get_supported_string():
    """
    Return supported versions as a string.
    """
    return "Mageia: " + ", ".join(sorted(version_to_config.keys()))



# Copyright (C) 2010,2011  Chris Lalancette <clalance@redhat.com>
# Copyright (C) 2012-2018  Chris Lalancette <clalancette@gmail.com>

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

"""
Fedora installation
"""

import os

import oz.OzException
import oz.RedHat
import oz.ozutil


class FedoraConfiguration(object):
    """
    The configuration class for Fedora.
    """
    def __init__(self, has_virtio_channel, use_yum, use_dev_cdrom_device,
                 createpart, directkernel, default_netdev, default_diskbus,
                 brokenisomethod, haverepo):
        self._has_virtio_channel = has_virtio_channel
        self._use_yum = use_yum
        self._use_dev_cdrom_device = use_dev_cdrom_device
        self._createpart = createpart
        self._directkernel = directkernel
        self._default_netdev = default_netdev
        self._default_diskbus = default_diskbus
        self._brokenisomethod = brokenisomethod
        self._haverepo = haverepo

    @property
    def has_virtio_channel(self):
        """
        Property method for whether this Fedora version has a virtio channel.
        """
        return self._has_virtio_channel

    @property
    def use_yum(self):
        """
        Property method for whether this Fedora version uses yum or dnf.
        """
        return self._use_yum

    @property
    def use_dev_cdrom_device(self):
        """
        Property method for whether this Fedora version uses /dev/cdrom on the
        kickstart command-line.
        """
        return self._use_dev_cdrom_device

    @property
    def createpart(self):
        """
        Property method for whether to create partitions before installation.
        """
        return self._createpart

    @property
    def directkernel(self):
        """
        Property method for whether this Fedora version supports direct kernel boot.
        """
        return self._directkernel

    @property
    def default_netdev(self):
        """
        Property method for the default netdev for this Fedora version.
        """
        return self._default_netdev

    @property
    def default_diskbus(self):
        """
        Property method for the default diskbus for this Fedora version.
        """
        return self._default_diskbus

    @property
    def brokenisomethod(self):
        """
        Property method for whether to add method to the anaconda install line.
        """
        return self._brokenisomethod

    @property
    def haverepo(self):
        """
        Property method for whether to use 'repo=' or 'method=' on the anaconda install line.
        """
        return self._haverepo


version_to_config = {
    '29': FedoraConfiguration(has_virtio_channel=True, use_yum=False,
                              use_dev_cdrom_device=True, createpart=False,
                              directkernel="cpio", default_netdev='virtio',
                              default_diskbus='virtio', brokenisomethod=False,
                              haverepo=True),
    '28': FedoraConfiguration(has_virtio_channel=True, use_yum=False,
                              use_dev_cdrom_device=True, createpart=False,
                              directkernel="cpio", default_netdev='virtio',
                              default_diskbus='virtio', brokenisomethod=False,
                              haverepo=True),
    '27': FedoraConfiguration(has_virtio_channel=True, use_yum=False,
                              use_dev_cdrom_device=True, createpart=False,
                              directkernel="cpio", default_netdev='virtio',
                              default_diskbus='virtio', brokenisomethod=False,
                              haverepo=True),
    '26': FedoraConfiguration(has_virtio_channel=True, use_yum=False,
                              use_dev_cdrom_device=True, createpart=False,
                              directkernel="cpio", default_netdev='virtio',
                              default_diskbus='virtio', brokenisomethod=False,
                              haverepo=True),
    '25': FedoraConfiguration(has_virtio_channel=True, use_yum=False,
                              use_dev_cdrom_device=True, createpart=False,
                              directkernel="cpio", default_netdev='virtio',
                              default_diskbus='virtio', brokenisomethod=False,
                              haverepo=True),
    '24': FedoraConfiguration(has_virtio_channel=True, use_yum=False,
                              use_dev_cdrom_device=True, createpart=False,
                              directkernel="cpio", default_netdev='virtio',
                              default_diskbus='virtio', brokenisomethod=False,
                              haverepo=True),
    '23': FedoraConfiguration(has_virtio_channel=True, use_yum=False,
                              use_dev_cdrom_device=True, createpart=False,
                              directkernel="cpio", default_netdev='virtio',
                              default_diskbus='virtio', brokenisomethod=False,
                              haverepo=True),
    '22': FedoraConfiguration(has_virtio_channel=True, use_yum=False,
                              use_dev_cdrom_device=True, createpart=False,
                              directkernel="cpio", default_netdev='virtio',
                              default_diskbus='virtio', brokenisomethod=False,
                              haverepo=True),
    '21': FedoraConfiguration(has_virtio_channel=True, use_yum=True,
                              use_dev_cdrom_device=True, createpart=False,
                              directkernel="cpio", default_netdev='virtio',
                              default_diskbus='virtio', brokenisomethod=False,
                              haverepo=True),
    '20': FedoraConfiguration(has_virtio_channel=True, use_yum=True,
                              use_dev_cdrom_device=True, createpart=False,
                              directkernel="cpio", default_netdev='virtio',
                              default_diskbus='virtio', brokenisomethod=False,
                              haverepo=True),
    '19': FedoraConfiguration(has_virtio_channel=True, use_yum=True,
                              use_dev_cdrom_device=True, createpart=False,
                              directkernel="cpio", default_netdev='virtio',
                              default_diskbus='virtio', brokenisomethod=False,
                              haverepo=True),
    '18': FedoraConfiguration(has_virtio_channel=True, use_yum=True,
                              use_dev_cdrom_device=True, createpart=False,
                              directkernel="cpio", default_netdev='virtio',
                              default_diskbus='virtio', brokenisomethod=True,
                              haverepo=True),
    '17': FedoraConfiguration(has_virtio_channel=True, use_yum=True,
                              use_dev_cdrom_device=True, createpart=False,
                              directkernel=None, default_netdev='virtio',
                              default_diskbus='virtio', brokenisomethod=True,
                              haverepo=True),
    '16': FedoraConfiguration(has_virtio_channel=True, use_yum=True,
                              use_dev_cdrom_device=False, createpart=False,
                              directkernel=None, default_netdev='virtio',
                              default_diskbus='virtio', brokenisomethod=True,
                              haverepo=True),
    '15': FedoraConfiguration(has_virtio_channel=True, use_yum=True,
                              use_dev_cdrom_device=False, createpart=False,
                              directkernel="cpio", default_netdev='virtio',
                              default_diskbus='virtio', brokenisomethod=True,
                              haverepo=True),
    '14': FedoraConfiguration(has_virtio_channel=True, use_yum=True,
                              use_dev_cdrom_device=False, createpart=False,
                              directkernel="cpio", default_netdev='virtio',
                              default_diskbus='virtio', brokenisomethod=True,
                              haverepo=True),
    '13': FedoraConfiguration(has_virtio_channel=False, use_yum=True,
                              use_dev_cdrom_device=False, createpart=False,
                              directkernel="cpio", default_netdev='virtio',
                              default_diskbus='virtio', brokenisomethod=True,
                              haverepo=True),
    '12': FedoraConfiguration(has_virtio_channel=False, use_yum=True,
                              use_dev_cdrom_device=False, createpart=True,
                              directkernel="cpio", default_netdev='virtio',
                              default_diskbus='virtio', brokenisomethod=True,
                              haverepo=True),
    '11': FedoraConfiguration(has_virtio_channel=False, use_yum=True,
                              use_dev_cdrom_device=False, createpart=True,
                              directkernel="cpio", default_netdev='virtio',
                              default_diskbus='virtio', brokenisomethod=True,
                              haverepo=True),
    '10': FedoraConfiguration(has_virtio_channel=False, use_yum=True,
                              use_dev_cdrom_device=False, createpart=False,
                              directkernel="cpio", default_netdev='virtio',
                              default_diskbus='virtio', brokenisomethod=True,
                              haverepo=True),
    '9': FedoraConfiguration(has_virtio_channel=False, use_yum=True,
                             use_dev_cdrom_device=False, createpart=False,
                             directkernel="cpio", default_netdev='virtio',
                             default_diskbus='virtio', brokenisomethod=False,
                             haverepo=True),
    '8': FedoraConfiguration(has_virtio_channel=False, use_yum=True,
                             use_dev_cdrom_device=False, createpart=False,
                             directkernel="cpio", default_netdev=None,
                             default_diskbus=None, brokenisomethod=False,
                             haverepo=False),
    '7': FedoraConfiguration(has_virtio_channel=False, use_yum=True,
                             use_dev_cdrom_device=False, createpart=False,
                             directkernel="cpio", default_netdev=None,
                             default_diskbus=None, brokenisomethod=False,
                             haverepo=False),
}


class FedoraGuest(oz.RedHat.RedHatLinuxCDYumGuest):
    """
    Class for Fedora 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, and 29 installation.
    """
    # Note that the 'brokenisomethod' and 'haverepo' parameters are completely
    # ignored now; we leave it in place for backwards API compatibility.
    def __init__(self, tdl, config, auto, nicmodel, haverepo, diskbus,  # pylint: disable=unused-argument
                 brokenisomethod, output_disk=None, macaddress=None,    # pylint: disable=unused-argument
                 assumed_update=None):
        self.config = version_to_config[tdl.update]
        if nicmodel is None:
            nicmodel = self.config.default_netdev
        if diskbus is None:
            diskbus = self.config.default_diskbus
        self.assumed_update = assumed_update

        # Prior to Fedora-22, we use yum; on F-22 and later, we use dnf.
        oz.RedHat.RedHatLinuxCDYumGuest.__init__(self, tdl, config, auto,
                                                 output_disk, nicmodel, diskbus,
                                                 True, True, self.config.directkernel,
                                                 macaddress, self.config.use_yum)

        if self.assumed_update is not None:
            self.log.warning("==== WARN: TDL contains Fedora update %s, which is newer than Oz knows about; pretending this is Fedora %s, but this may fail ====", tdl.update, assumed_update)

        if self.config.has_virtio_channel:
            self.virtio_channel_name = 'org.fedoraproject.anaconda.log.0'

    def _modify_iso(self):
        """
        Method to modify the ISO for autoinstallation.
        """
        self._copy_kickstart(os.path.join(self.iso_contents, "ks.cfg"))

        if self.config.use_dev_cdrom_device:
            initrdline = "  append initrd=initrd.img ks=cdrom:/dev/cdrom:/ks.cfg"
        else:
            initrdline = "  append initrd=initrd.img ks=cdrom:/ks.cfg"
        if self.tdl.installtype == "url":
            if self.config.haverepo:
                initrdline += " repo="
            else:
                initrdline += " method="
            initrdline += self.url
        else:
            # if the installtype is iso, then due to a bug in anaconda we leave
            # out the method completely
            if not self.config.brokenisomethod:
                initrdline += " method=cdrom:/dev/cdrom"
        self._modify_isolinux(initrdline)

    def generate_diskimage(self, size=10, force=False):
        """
        Method to generate a diskimage.  By default, a blank diskimage of
        10GB will be created; the caller can override this with the size
        parameter, specified in GB.  If force is False (the default), then
        a diskimage will not be created if a cached JEOS is found.  If
        force is True, a diskimage will be created regardless of whether a
        cached JEOS exists.  See the oz-install man page for more
        information about JEOS caching.
        """
        # If given a blank diskimage, Fedora 11/12 stops very early in
        # install with a message about losing all of your data on the
        # drive (it differs between them).
        #
        # To avoid that message, just create a partition table that spans
        # the entire disk
        return self._internal_generate_diskimage(size, force, self.config.createpart)

    def get_auto_path(self):
        """
        Method to create the correct path to the Fedora kickstart files.
        """
        # If we are doing our best with an unknown Fedora update, use the
        # newest known auto file; otherwise, do the usual thing.
        if self.assumed_update is not None:
            return oz.ozutil.generate_full_auto_path(self.tdl.distro + self.assumed_update + ".auto")
        return oz.ozutil.generate_full_auto_path(self.tdl.distro + self.tdl.update + ".auto")


def get_class(tdl, config, auto, output_disk=None, netdev=None, diskbus=None,
              macaddress=None):
    """
    Factory method for Fedora installs.
    """

    newest = sorted(version_to_config.keys(), key=int)[-1]
    if tdl.update == 'rawhide' or int(tdl.update) > int(newest):
        return FedoraGuest(tdl, config, auto, netdev, True, diskbus, False,
                           output_disk, macaddress, newest)

    if tdl.update in version_to_config.keys():
        return FedoraGuest(tdl, config, auto, netdev, True, diskbus, False,
                           output_disk, macaddress, None)


def get_supported_string():
    """
    Return supported versions as a string.
    """
    return "Fedora: " + ", ".join(sorted(version_to_config.keys(), key=int))



# Copyright (C) 2010,2011  Chris Lalancette <clalance@redhat.com>
# Copyright (C) 2012-2018  Chris Lalancette <clalancette@gmail.com>

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

"""
Windows installation
"""

import os
import random
import re
import shutil

import lxml.etree

import oz.Guest
import oz.OzException
import oz.ozutil


class Windows(oz.Guest.CDGuest):
    """
    Shared Windows base class.
    """
    def __init__(self, tdl, config, auto, output_disk, netdev, diskbus,
                 macaddress):
        oz.Guest.CDGuest.__init__(self, tdl, config, auto, output_disk,
                                  netdev, "localtime", "usb", diskbus, True,
                                  False, macaddress)

        if self.tdl.key is None:
            raise oz.OzException.OzException("A key is required when installing Windows")


class Windows_v5(Windows):
    """
    Class for Windows versions based on kernel 5.x (2000, XP, and 2003).
    """
    def __init__(self, tdl, config, auto, output_disk, netdev, diskbus,
                 macaddress):
        Windows.__init__(self, tdl, config, auto, output_disk, netdev, diskbus,
                         macaddress)

        if self.tdl.update == "2000" and self.tdl.arch != "i386":
            raise oz.OzException.OzException("Windows 2000 only supports i386 architecture")

        self.winarch = self.tdl.arch
        if self.winarch == "x86_64":
            self.winarch = "amd64"

    def _generate_new_iso(self):
        """
        Method to create a new ISO based on the modified CD/DVD.
        """
        self.log.debug("Generating new ISO")
        oz.ozutil.subprocess_check_output(["genisoimage",
                                           "-b", "cdboot/boot.bin",
                                           "-no-emul-boot", "-boot-load-seg",
                                           "1984", "-boot-load-size", "4",
                                           "-iso-level", "2", "-J", "-l", "-D",
                                           "-N", "-joliet-long",
                                           "-relaxed-filenames", "-v",
                                           "-V", "Custom",
                                           "-o", self.output_iso,
                                           self.iso_contents],
                                          printfn=self.log.debug)

    def generate_diskimage(self, size=10, force=False):
        """
        Method to generate a diskimage.  By default, a blank diskimage of
        10GB will be created; the caller can override this with the size
        parameter, specified in GB.  If force is False (the default), then
        a diskimage will not be created if a cached JEOS is found.  If
        force is True, a diskimage will be created regardless of whether a
        cached JEOS exists.  See the oz-install man page for more
        information about JEOS caching.
        """
        createpart = False
        if self.tdl.update == "2000":
            # If given a blank diskimage, windows 2000 stops very early in
            # install with a message:
            #
            #  Setup has determined that your computer's startup hard disk is
            #  new or has been erased...
            #
            # To avoid that message, create a partition table that spans
            # the entire disk
            createpart = True
        return self._internal_generate_diskimage(size, force, createpart)

    def _modify_iso(self):
        """
        Method to make the boot ISO auto-boot with appropriate parameters.
        """
        self.log.debug("Modifying ISO")

        os.mkdir(os.path.join(self.iso_contents, "cdboot"))
        self._geteltorito(self.orig_iso, os.path.join(self.iso_contents,
                                                      "cdboot", "boot.bin"))

        outname = os.path.join(self.iso_contents, self.winarch, "winnt.sif")

        if self.default_auto_file():
            # if this is the oz default siffile, we modify certain parameters
            # to make installation succeed
            computername = "OZ" + str(random.randrange(1, 900000))

            def _sifsub(line):
                """
                Method that is called back from oz.ozutil.copy_modify_file() to
                modify sif files as appropriate for Windows 2000/XP/2003.
                """
                if re.match(" *ProductKey", line):
                    return "    ProductKey=" + self.tdl.key + "\n"
                elif re.match(" *ProductID", line):
                    return "    ProductID=" + self.tdl.key + "\n"
                elif re.match(" *ComputerName", line):
                    return "    ComputerName=" + computername + "\n"
                elif re.match(" *AdminPassword", line):
                    return "    AdminPassword=" + self.rootpw + "\n"
                return line

            oz.ozutil.copy_modify_file(self.auto, outname, _sifsub)
        else:
            # if the user provided their own siffile, do not override their
            # choices; the user gets to keep both pieces if something breaks
            shutil.copy(self.auto, outname)

    def install(self, timeout=None, force=False):
        """
        Method to run the operating system installation.
        """
        internal_timeout = timeout
        if internal_timeout is None:
            internal_timeout = 3600
        return self._do_install(internal_timeout, force, 1)


class Windows_v6(Windows):
    """
    Class for Windows versions based on kernel 6.x (2008, 7, 2012, 8, and 8.1).
    """
    def __init__(self, tdl, config, auto, output_disk, netdev, diskbus,
                 macaddress):
        Windows.__init__(self, tdl, config, auto, output_disk, netdev, diskbus,
                         macaddress)

        self.winarch = "x86"
        if self.tdl.arch == "x86_64":
            self.winarch = "amd64"

    def _generate_new_iso(self):
        """
        Method to create a new ISO based on the modified CD/DVD.
        """
        self.log.debug("Generating new ISO")
        # NOTE: Windows 2008 is very picky about which arguments to genisoimage
        # will generate a bootable CD, so modify these at your own risk
        oz.ozutil.subprocess_check_output(["genisoimage",
                                           "-b", "cdboot/boot.bin",
                                           "-no-emul-boot", "-c", "BOOT.CAT",
                                           "-iso-level", "2", "-J", "-l", "-D",
                                           "-N", "-joliet-long",
                                           "-relaxed-filenames", "-v",
                                           "-V", "Custom", "-udf",
                                           "-o", self.output_iso,
                                           self.iso_contents],
                                          printfn=self.log.debug)

    def _modify_iso(self):
        """
        Method to make the boot ISO auto-boot with appropriate parameters.
        """
        self.log.debug("Modifying ISO")

        os.mkdir(os.path.join(self.iso_contents, "cdboot"))
        self._geteltorito(self.orig_iso, os.path.join(self.iso_contents,
                                                      "cdboot", "boot.bin"))

        outname = os.path.join(self.iso_contents, "autounattend.xml")

        if self.default_auto_file():
            # if this is the oz default unattend file, we modify certain
            # parameters to make installation succeed
            doc = lxml.etree.parse(self.auto)

            for component in doc.xpath('/ms:unattend/ms:settings/ms:component',
                                       namespaces={'ms': 'urn:schemas-microsoft-com:unattend'}):
                component.set('processorArchitecture', self.winarch)

            keys = doc.xpath('/ms:unattend/ms:settings/ms:component/ms:ProductKey',
                             namespaces={'ms': 'urn:schemas-microsoft-com:unattend'})
            if len(keys) != 1:
                raise oz.OzException.OzException("Invalid autounattend file; expected 1 key, saw %d" % (len(keys)))
            keys[0].text = self.tdl.key

            adminpw = doc.xpath('/ms:unattend/ms:settings/ms:component/ms:UserAccounts/ms:AdministratorPassword/ms:Value',
                                namespaces={'ms': 'urn:schemas-microsoft-com:unattend'})
            if len(adminpw) != 1:
                raise oz.OzException.OzException("Invalid autounattend file; expected 1 admin password, saw %d" % (len(adminpw)))
            adminpw[0].text = self.rootpw

            autologinpw = doc.xpath('/ms:unattend/ms:settings/ms:component/ms:AutoLogon/ms:Password/ms:Value',
                                    namespaces={'ms': 'urn:schemas-microsoft-com:unattend'})
            if len(autologinpw) != 1:
                raise oz.OzException.OzException("Invalid autounattend file; expected 1 auto logon password, saw %d" % (len(autologinpw)))
            autologinpw[0].text = self.rootpw

            with open(outname, 'w') as f:
                f.write(lxml.etree.tostring(doc, pretty_print=True, encoding="unicode"))
        else:
            # if the user provided their own unattend file, do not override
            # their choices; the user gets to keep both pieces if something
            # breaks
            shutil.copy(self.auto, outname)

    def install(self, timeout=None, force=False):
        internal_timeout = timeout
        if internal_timeout is None:
            internal_timeout = 8500
        return self._do_install(internal_timeout, force, 2)


class Windows_v10(Windows):
    """
    Class for Windows versions based on kernel 10.x (2016 and 10).
    """
    def __init__(self, tdl, config, auto, output_disk, netdev, diskbus,
                 macaddress):
        Windows.__init__(self, tdl, config, auto, output_disk, netdev, diskbus,
                         macaddress)

        self.winarch = "x86"
        if self.tdl.arch == "x86_64":
            self.winarch = "amd64"

    def _generate_new_iso(self):
        """
        Method to create a new ISO based on the modified CD/DVD.
        """
        self.log.debug("Generating new ISO")
        oz.ozutil.subprocess_check_output(["genisoimage",
                                           "-b", "cdboot/boot.bin",
                                           "-no-emul-boot", "-c", "BOOT.CAT",
                                           "-iso-level", "2", "-J", "-l", "-D",
                                           "-N", "-joliet-long", "-allow-limited-size",
                                           "-relaxed-filenames", "-v",
                                           "-V", "Custom", "-udf",
                                           "-o", self.output_iso,
                                           self.iso_contents],
                                          printfn=self.log.debug)

    def _modify_iso(self):
        """
        Method to make the boot ISO auto-boot with appropriate parameters.
        """
        self.log.debug("Modifying ISO")

        os.mkdir(os.path.join(self.iso_contents, "cdboot"))
        self._geteltorito(self.orig_iso, os.path.join(self.iso_contents,
                                                      "cdboot", "boot.bin"))

        outname = os.path.join(self.iso_contents, "autounattend.xml")

        if self.default_auto_file():
            # if this is the oz default unattend file, we modify certain
            # parameters to make installation succeed
            doc = lxml.etree.parse(self.auto)

            for component in doc.xpath('/ms:unattend/ms:settings/ms:component',
                                       namespaces={'ms': 'urn:schemas-microsoft-com:unattend'}):
                component.set('processorArchitecture', self.winarch)

            keys = doc.xpath('/ms:unattend/ms:settings/ms:component/ms:ProductKey',
                             namespaces={'ms': 'urn:schemas-microsoft-com:unattend'})
            if len(keys) != 1:
                raise oz.OzException.OzException("Invalid autounattend file; expected 1 key, saw %d" % (len(keys)))
            keys[0].text = self.tdl.key

            adminpw = doc.xpath('/ms:unattend/ms:settings/ms:component/ms:UserAccounts/ms:AdministratorPassword/ms:Value',
                                namespaces={'ms': 'urn:schemas-microsoft-com:unattend'})
            if len(adminpw) != 1:
                raise oz.OzException.OzException("Invalid autounattend file; expected 1 admin password, saw %d" % (len(adminpw)))
            adminpw[0].text = self.rootpw

            autologinpw = doc.xpath('/ms:unattend/ms:settings/ms:component/ms:AutoLogon/ms:Password/ms:Value',
                                    namespaces={'ms': 'urn:schemas-microsoft-com:unattend'})
            if len(autologinpw) != 1:
                raise oz.OzException.OzException("Invalid autounattend file; expected 1 auto logon password, saw %d" % (len(autologinpw)))
            autologinpw[0].text = self.rootpw

            with open(outname, 'w') as f:
                f.write(lxml.etree.tostring(doc, pretty_print=True, encoding="unicode"))
        else:
            # if the user provided their own unattend file, do not override
            # their choices; the user gets to keep both pieces if something
            # breaks
            shutil.copy(self.auto, outname)

    def install(self, timeout=None, force=False):
        internal_timeout = timeout
        if internal_timeout is None:
            internal_timeout = 8500
        return self._do_install(internal_timeout, force, 2)


def get_class(tdl, config, auto, output_disk=None, netdev=None, diskbus=None,
              macaddress=None):
    """
    Factory method for Windows installs.
    """
    if tdl.update in ["2000", "XP", "2003"]:
        return Windows_v5(tdl, config, auto, output_disk, netdev,
                          diskbus, macaddress)
    if tdl.update in ["2008", "7", "2012", "8", "8.1"]:
        return Windows_v6(tdl, config, auto, output_disk, netdev, diskbus,
                          macaddress)
    if tdl.update in ["2016", "10"]:
        return Windows_v10(tdl, config, auto, output_disk, netdev, diskbus,
                           macaddress)


def get_supported_string():
    """
    Return supported versions as a string.
    """
    return "Windows: 2000, XP, 2003, 7, 2008, 2012, 8, 8.1, 2016, 10"



"""
Class for automated operating system installation.

Oz is a set of classes to do automated operating system installation.  It
has built-in knowledge of the proper things to do for each of the supported
operating systems, so the data that the user must provide is very minimal.
This data is supplied in the form of an XML document that describes what
type of operating system is to be installed and where to get the
installation media.  Oz handles the rest.

The simplest Oz program (without error handling or any advanced features)
would look something like:

import oz.TDL
import oz.GuestFactory

tdl_xml = \"\"\"
<template>
  <name>f13jeos</name>
  <os>
    <name>Fedora</name>
    <version>13</version>
    <arch>x86_64</arch>
    <install type='url'>
      <url>http://download.fedoraproject.org/pub/fedora/linux/releases/13/Fedora/x86_64/os/</url>
    </install>
  </os>
  <description>Fedora 13</description>
</template>
\"\"\"

tdl = oz.TDL.TDL(tdl_xml)
guest = oz.GuestFactory.guest_factory(tdl, None, None)
guest.generate_install_media()
guest.generate_diskimage()
guest.install()
"""
